import importlib
import os.path

import numpy as np
from lxml import etree
from lxml.builder import E


def load_walls(maze_array):
    horizontal_walls = []
    vertical_walls = []
    maze_array_np = np.array(maze_array)
    while np.isin(1, maze_array_np):
        first_tile = np.flip(np.argwhere(maze_array_np == 1)[0].copy()).tolist()

        # build the horizontal wall
        horizontal_wall = [first_tile]
        current_x_coordinates = first_tile[0]  # Add wall parts at the left
        while current_x_coordinates - 1 > 0 and maze_array_np[first_tile[1]][current_x_coordinates - 1].item() == 1:
            horizontal_wall.insert(0, [current_x_coordinates - 1, first_tile[1]])
            current_x_coordinates -= 1
        current_x_coordinates = first_tile[0]  # Add wall parts at the right
        while current_x_coordinates + 1 < len(maze_array[0]) \
                and maze_array_np[first_tile[1]][current_x_coordinates + 1].item() == 1:
            horizontal_wall.append([current_x_coordinates + 1, first_tile[1]])
            current_x_coordinates += 1

        # build the vertical wall
        vertical_wall = [first_tile]
        current_y_coordinates = first_tile[1]  # Add wall parts at the left
        while current_y_coordinates - 1 > 0 and maze_array_np[current_y_coordinates - 1][first_tile[0]].item() == 1:
            horizontal_wall.insert(0, [first_tile[1], current_y_coordinates - 1])
            current_y_coordinates -= 1
        current_y_coordinates = first_tile[1]  # Add wall parts at the right
        while current_y_coordinates + 1 < len(maze_array) \
                and maze_array_np[current_y_coordinates + 1][first_tile[0]].item() == 1:
            vertical_wall.append([first_tile[0], current_y_coordinates + 1])
            current_y_coordinates += 1

        # Choose the longer wall and add it to walls list.
        if len(horizontal_wall) > len(vertical_wall):
            # replace every tile in the chosen wall by 0, so it will not be added again
            row_index = horizontal_wall[0][1]
            col_from, col_to = horizontal_wall[0][0], horizontal_wall[-1][0]
            maze_array_np[row_index, col_from:col_to + 1] = 0

            # Add this wall to the memory
            horizontal_walls.append([horizontal_wall[0], horizontal_wall[-1]])
        else:
            # replace every tile in the chosen wall by 0, so it will not be added again
            col_index = vertical_wall[0][0]
            row_from, row_to = vertical_wall[0][1], vertical_wall[-1][1]
            maze_array_np[row_from:row_to + 1, col_index] = 0

            # Add this wall to the memory
            vertical_walls.append([vertical_wall[0], vertical_wall[-1]])
    return horizontal_walls, vertical_walls


def generate_xml(map_name: str) -> (dict, str):
    """
    Generate an ant-maze environment model from a base model and maze walls description.
    :returns: (as a tuple of two elements)
     - A 'dict' that contains maze description file information.
     - A 'str' that correspond to the path where the resultant xml file has been generated.
    """

    # Load base ant-maze file etree:

    # Get the path to the current directory.
    current_directory = os.path.dirname(__file__)
    tree = etree.parse(current_directory + "/ant_maze.xml")

    # Find 'world_body' int ant-maze xml file:
    world_body_node = None
    for child in tree.getroot():
        if child.tag == "worldbody":
            world_body_node = child

    # Load maps data
    maze_array = np.array(importlib.import_module("rlnav.ant_maze.maps." + map_name).maze_array)

    # Load walls
    width = len(maze_array[0])
    height = len(maze_array)
    horizontal_walls, vertical_walls = load_walls(maze_array)

    # Build walls xml nodes using loaded walls data
    for wall_id, wall in enumerate(horizontal_walls):
        start_x = wall[0][0] - width / 2 + 0.5
        end_x = wall[1][0] - width / 2 + 0.5
        position_x = (end_x + start_x) / 2
        position_y = - (wall[0][1] - height / 2 + 0.5)
        position = str(position_x) + " " + str(position_y) + " 0.5"
        wall_size = str((end_x - start_x + 1) / 2) + " 0.5 0.5"
        node = E.body(
            E.geom(type="box", size=wall_size, contype="1", conaffinity="1", rgba="0.4 0.4 0.4 1"),
            name="horizontal_wall_" + str(wall_id), pos=position
        )
        world_body_node.append(node)

    for wall_id, wall in enumerate(vertical_walls):
        start_y = - (wall[1][1] - height / 2 + 0.5)
        end_y = - (wall[0][1] - height / 2 + 0.5)
        position_x = wall[0][0] - width / 2 + 0.5
        position_y = (end_y + start_y) / 2
        position = str(position_x) + " " + str(position_y) + " 0.5"
        wall_size = "0.5 " + str((end_y - start_y + 1) / 2) + " 0.5"
        node = E.body(
            E.geom(type="box", size=wall_size, contype="1", conaffinity="1", rgba="0.4 0.4 0.4 1"),
            name="vertical_wall_" + str(wall_id), pos=position
        )
        world_body_node.append(node)

    # Modify the position of the global camera, that is supposed to view the entire maze.
    # Its position then depends on the width of the environment.
    camera = world_body_node.find(".//camera[@name='env_view']")  # Find camera by name
    if camera is not None:
        new_z_position = max(maze_array.shape[:2])
        camera.set('pos', '0 0 ' + str(new_z_position))  # Update position as an example
        # camera.set('xyaxes', '1 0 0 0 0 -1')  # Flip the camera orientation to make it right-side up

        # # 4. Save the modified XML back to a new file
        # tree.write("modified_model.xml", pretty_print=True)
    else:
        print("Camera not found in the XML.")

    xml_output_path = current_directory + "/generated/" + "ant_maze_" + map_name + ".xml"
    tree.write(xml_output_path)

    # Make generated file more pretty
    parser = etree.XMLParser(remove_blank_text=True)
    tree = etree.parse(xml_output_path, parser)
    tree.write(xml_output_path, pretty_print=True)

    return maze_array, xml_output_path


if __name__ == "__main__":
    # maps = ["empty_room", "extreme_maze", "four_rooms", "hard_maze", "medium_maze"]
    # for map_name in maps:
    #     generate_xml(map_name)
    map_name = "four_rooms_small"
    generate_xml(map_name)



import os
import random

import numpy as np
import importlib
from omg import *

def get_position(w, h, block_size):
    x, y = w * block_size, h * block_size
    x += int(block_size / 2)
    y += int(block_size / 2)
    return x, y


def build_wall(map_array, random_walls_textures=False, block_size=96, discrete_goals_set=False):
    things = []
    linedefs = []
    vertexes = []
    v_indexes = {}

    max_w = len(map_array[0]) - 1
    max_h = len(map_array) - 1

    def __is_edge(w, h):
        return w in (0, max_w) or h in (0, max_h)

    def __add_start(w, h):
        x, y = get_position(w, max_h - h, block_size)
        things.append(ZThing(*[len(things) + 1000, x, y, 0, 0, 9001, 22279]))

    def __add_vertex(w, h):
        if (w, h) in v_indexes:
            return
        x, y = get_position(w, max_h - h, block_size)
        print("adding vertex at ", [x, y])
        v_indexes[w, h] = len(vertexes)
        vertexes.append(Vertex(x, y))

    def __add_line(start, end, edge=False):
        assert start in v_indexes
        assert end in v_indexes

        mask = 1
        if random_walls_textures:
            left = random.randint(0, 66)
            right = random.randint(0, 66)
        else:
            left = right = 0
        if __is_edge(*start) and __is_edge(*end):
            if not edge:
                return
            else:
                # Changed the right side (one towards outside the map)
                # to be -1 (65535 for Doom)
                right = 65535
                mask = 15

        # Flipped end and start vertices to make lines "point" at right direction (mostly to see if it works)
        line_properties = [v_indexes[end], v_indexes[start], mask
                           ] + [0] * 6 + [left, right]
        line = ZLinedef(*line_properties)
        linedefs.append(line)

    for h, row in enumerate(map_array):
        for w, block in enumerate(row):
            if block == 1:
                __add_vertex(w, h)
            else:
                pass

    corners = [(0, 0), (max_w, 0), (max_w, max_h), (0, max_h)]

    for i in range(len(corners)):
        if i != len(corners) - 1:
            p_1 = corners[i + 1]
            p_2 = corners[i]
        else:
            p_1 = corners[0]
            p_2 = corners[i]

        if p_1[1] > p_2[1]:
            for h in reversed(range(p_2[1], p_1[1])):
                __add_line((p_1[0], h + 1), (p_2[0], h), True)
        elif p_1[1] < p_2[1]:
            for h in range(p_1[1], p_2[1]):
                __add_line((p_1[0], h), (p_2[0], h + 1), True)
        elif p_1[0] > p_2[0]:
            for w in reversed(range(p_2[0], p_1[0])):
                __add_line((w + 1, p_1[1]), (w, p_2[1]), True)
        elif p_1[0] < p_2[0]:
            for w in range(p_1[0], p_2[0]):
                __add_line((w, p_1[1]), (w + 1, p_2[1]), True)

    # Now connect the walls
    for h, row in enumerate(map_array):
        for w, _ in enumerate(row):
            if (w, h) not in v_indexes:
                __add_start(w, h)
                continue

            if (w + 1, h) in v_indexes:
                __add_line((w, h), (w + 1, h))

            if (w, h + 1) in v_indexes:
                __add_line((w, h), (w, h + 1))

    # Place goals items
    if discrete_goals_set:
        goals_positions = np.where(map_array == 3)
        goals_types = [32, 35, 30, 41, 36, 70, 56, 55, 47, 44]
        for goal_id in range(len(goals_positions[0])):
            h = goals_positions[0][goal_id]
            w = goals_positions[1][goal_id]
            goal_type = goals_types[goal_id % len(goals_types)]

            x, y = get_position(w, max_h - h, block_size)
            things.append(ZThing(0, x, y, 0, 0, goal_type, 263))
    else:
        candidates = np.where(map_array == 2)
        position_index = random.randint(0, len(candidates[0]) - 1)
        h = 1
        w = 3
        x, y = get_position(w, max_h - h, block_size)
        things.append(ZThing(0, x, y, 0, 0, 32, 263))  # type 32 = red pillar

    return things, vertexes, linedefs

def build_wad(output_directory: str, map_array,
              random_walls_textures=True,
              discrete_goals_set=False,
              behaviour_file="static_goal.acs", block_size=96):

    if not output_directory.endswith("/"): output_directory += "/"
    assert os.path.isdir(output_directory)

    # Verify inputs
    ## behaviour_file
    configurations_files_path = os.path.dirname(os.path.abspath(__file__)) + "/configurations/"
    if not "." in behaviour_file:
        behaviour_file += ".acs"
    if behaviour_file.split(".")[-1] != "acs":
        raise ValueError("Wrong behaviour file extension '." + behaviour_file.split(".")[-1] + "'")
    behaviour_file_path = configurations_files_path + behaviour_file
    if not os.path.exists(behaviour_file_path):
        raise ValueError("File " + behaviour_file + "not found in directory " + configurations_files_path)

    ## map_array
    assert isinstance(map_array, np.ndarray)

    # Build wad file
    new_wad = WAD()
    new_map = MapEditor()
    new_map.Linedef = ZLinedef
    new_map.Thing = ZThing
    new_map.behavior = Lump(from_file=behaviour_file_path or None)
    new_map.scripts = Lump(from_file=None)

    new_map.sectors = [Sector(0, 128, 'CEIL5_2', 'CEIL5_2', 240, 0, 0)]
    # new_map.sectors = [Sector(0, 128, 'ZIMMER8', 'ZZWOLF1', 240, 0, 0)]
    things, vertexes, linedefs = build_wall(map_array,
                                            random_walls_textures=random_walls_textures,
                                            discrete_goals_set=discrete_goals_set,
                                            block_size=block_size)

    candidates = np.where(map_array == 2)
    cand_index = random.randint(0, len(candidates[0]) - 1)
    w = candidates[0][cand_index]
    h = candidates[0][cand_index]
    x, y = get_position(w, len(map_array) - h - 1, block_size)
    new_map.things = things + [ZThing(0, x, y, 0, 0, 1, 7)]

    new_map.vertexes = vertexes
    new_map.linedefs = linedefs

    texture_names = ['ZZWOLF6', 'ZIMMER8', 'TEKBRON1', 'WOOD5', 'WOODMET3', 'ZDOORB1', 'ZZWOLF9', 'TEKGREN1',
    'WOOD6', 'WOODMET4', 'TEKLITE', 'TEKWALL1', 'WOODMET2', 'TANROCK4', 'ZZWOLF10', 'WOOD9', 'ZZZFACE6', 'TANROCK3',
    'ZZZFACE2', 'ZZZFACE1', 'ZZWOLF11', 'TEKGREN4', 'ZZWOLF4', 'TEKBRON2', 'TEKGREN3', 'WOOD3', 'ZZZFACE4', 'WOOD1',
    'ZDOORF1', 'ZIMMER4', 'ZZZFACE5', 'TEKWALL6', 'WOOD4', 'TANROCK2', 'TEKGREN2', 'ZIMMER7', 'TANROCK8', 'ZZWOLF7',
    'ZZZFACE8', 'ZIMMER5', 'WOODSKUL', 'ZZWOLF3', 'ZZWOLF2', 'ZZZFACE9', 'TEKWALL2', 'ZELDOOR', 'ZIMMER2', 'WOOD8',
    'TEKWALL5', 'TANROCK5', 'TEKWALL4', 'TEKLITE2', 'ZZWOLF12', 'WOODMET1', 'ZZZFACE3', 'ZIMMER3', 'ZZWOLF1',
    'TANROCK7', 'WOODGARG', 'ZZWOLF13', 'WOOD12', 'ZZZFACE7', 'TEKGREN5', 'WOODVERT', 'ZZWOLF5', 'WOOD7',
    'ZIMMER1']

    new_map.sidedefs = [Sidedef(0, 0, '-', '-', texture_name, 0) for texture_name in texture_names]
    new_map.sidedefs.append(Sidedef(0, 0, '-', '-', '-', 0))

    new_wad.maps['MAP00'] = new_map.to_lumps()
    new_wad.to_file(output_directory + "/map" + ".wad")
